/*
 *	vim: ts=4 shiftwidth=4
 *	$Abso: abso/web/gestion/sql/C/triggers.c,v 1.2 2002/08/28 14:09:10 mat Exp $
 */

#include <executor/spi.h>		/* this is what you need to work with SPI */
#include <commands/trigger.h>	/* -"- and triggers */

#ifdef DBG
#define D(x...) { elog(NOTICE, ##x); }
#else
#define D(x...) { };
#endif

#define CONTENUMAX 4096
#define QUERYMAX (CONTENUMAX + 256)
#define QUERY "INSERT INTO historique (operation, \"table\", contenu) VALUES (%s, %s, %s)"

extern Datum historisation(PG_FUNCTION_ARGS);
static void append_contenu(char **, char *, char *);
static void append_contenu_new(char **, char *, char *, char *);
static char *quote_it(char *);

PG_FUNCTION_INFO_V1(historisation);

Datum
historisation(PG_FUNCTION_ARGS)
{
	TriggerData *trigdata = (TriggerData *)fcinfo->context;
	char	   *relname;		/* triggered relation name */
	Relation	rel;			/* triggered relation */
	HeapTuple	rettuple = NULL;
	TupleDesc	tupdesc;		/* tuple description */

	int			i;
	char	   *attrname,
			   *attrval,
			   *attrval2,
			   *query,
			   *type,
			   *contenu;

	if (!CALLED_AS_TRIGGER(fcinfo))
		elog(ERROR, "historisation: not fired by trigger manager");
	if (TRIGGER_FIRED_FOR_STATEMENT(trigdata->tg_event))
		elog(ERROR, "historisation: can't process STATEMENT events");
	if (TRIGGER_FIRED_BEFORE(trigdata->tg_event))
		elog(ERROR, "historisation: must be fired after event");

	rel = trigdata->tg_relation;
	relname = SPI_getrelname(rel);
	tupdesc = rel->rd_att;

	contenu = (char *)palloc(CONTENUMAX + 1);
	bzero(contenu, CONTENUMAX + 1);

	if (TRIGGER_FIRED_BY_INSERT(trigdata->tg_event))
	{
		type = strdup("INSERT");
		rettuple = trigdata->tg_trigtuple;

		for (i = 1; (attrname = SPI_fname(tupdesc, i)) != NULL; i++)
		{
			attrval = SPI_getvalue(rettuple, tupdesc, i);
			append_contenu(&contenu, attrname, attrval);
			if (attrval != NULL)
				pfree(attrval);
			pfree(attrname);
		}
	}
	else if (TRIGGER_FIRED_BY_UPDATE(trigdata->tg_event))
	{
		type = strdup("UPDATE");
		rettuple = trigdata->tg_newtuple;
		append_contenu(&contenu, SPI_fname(tupdesc, 1), SPI_getvalue(rettuple, tupdesc, 1));
		for (i = 2; (attrname = SPI_fname(tupdesc, i)) != NULL; i++)
		{
			attrval = SPI_getvalue(rettuple, tupdesc, i);
			attrval2 = SPI_getvalue(trigdata->tg_trigtuple, tupdesc, i);
			if (strcmp(attrval, attrval2) != 0)
				append_contenu_new(&contenu, attrname, attrval2, attrval);
			else
				append_contenu(&contenu, attrname, attrval);
			if (attrval2 != NULL)
				pfree(attrval2);
			if (attrval != NULL)
				pfree(attrval);
			pfree(attrname);
		}
	}
	else
	{
		type = strdup("DELETE");
		rettuple = trigdata->tg_trigtuple;
		for (i = 1; (attrname = SPI_fname(tupdesc, i)) != NULL; i++)
		{
			attrval = SPI_getvalue(rettuple, tupdesc, i);
			append_contenu(&contenu, attrname, attrval);
			if (attrval != NULL)
				pfree(attrval);
			pfree(attrname);
		}
	}

	/* on enleve le ", " qui se trouve a la fin */
	*(contenu + strlen(contenu) - 2) = '\0';

	SPI_connect();
	query = (char *)palloc(QUERYMAX);
	snprintf(query, QUERYMAX, QUERY, quote_it(type), quote_it(relname), quote_it(contenu));
	SPI_exec(query, 1);
	SPI_finish();

	pfree(contenu);
	pfree(relname);

	return PointerGetDatum(rettuple);
}


static char *
quote_it(char *lptr)
{
	char	   *result = NULL,
			   *d = NULL,
			   *c;
	int			len = 0;

	if (lptr == NULL)
		result = strdup("NULL");
	else
	{
		len = strlen(lptr);
		/* 3 : ' du dbut et fin + \0 */
		d = (char *)palloc(len * 2 + 3);
		result = d;
		bzero(d, len * 2 + 3);
		c = lptr;

		*d++ = '\'';
		while (len-- > 0)
		{
			if (*lptr == '\'')
				*d++ = '\'';
			if (*lptr == '\\')
				*d++ = '\\';
			*d++ = *lptr++;
		}
		*d++ = '\'';
		*d = '\0';
	}

	return result;
}

static void
append_contenu(char **contenu, char *qui, char *quoi)
{
	strncat(*contenu, qui, CONTENUMAX);
	strncat(*contenu, "=", CONTENUMAX);
	strncat(*contenu, quote_it(quoi), CONTENUMAX);
	strncat(*contenu, ", ", CONTENUMAX);
}

static void
append_contenu_new(char **contenu, char *qui, char *quoi, char *new)
{
	strncat(*contenu, qui, CONTENUMAX);
	strncat(*contenu, " ", CONTENUMAX);
	strncat(*contenu, quote_it(quoi), CONTENUMAX);
	strncat(*contenu, " -> ", CONTENUMAX);
	strncat(*contenu, quote_it(new), CONTENUMAX);
	strncat(*contenu, ", ", CONTENUMAX);
}
